(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2009                                               *)
(*    CEA (Commissariat  l'nergie Atomique)                             *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* $Id: journal.ml,v 1.32 2009-01-28 14:34:54 uid568 Exp $ *)

(* ****************************************************************************)
(* ****************************************************************************)
(* ****************************************************************************)

(* Disclaimer
   ----------
   This module uses very unsafe caml features (module Obj).
   Modify it at your own risk.
   Sometimes the caml type system does not help you here.
   Introducing a bug here may introduce some "segmentation faults" in Frama-C *)

(* ****************************************************************************)
(* ****************************************************************************)
(* ****************************************************************************)

include Log.Register
  (struct
     let channel = Log.kernel_channel_name
     let label = Log.kernel_label_name
     let verbose_atleast n = Cmdline.verbose_level >= n
     let debug_atleast n = Cmdline.debug_level >= n
   end)

(** Journalization of functions *)

(* ****************************************************************************)
(** {2 Journal management} *)
(* ****************************************************************************)

exception LoadingError of string

let started = ref true
let enable_exn = ref true
let before_run = ref true

let journal_buffer = Buffer.create 8192
let journal_fmt = Format.formatter_of_buffer journal_buffer

(* formatter settings *)
let line_length = 78
let () = Format.pp_set_margin journal_fmt line_length

let something_to_write = ref false

let now () = Unix.localtime (Unix.time ())

let filename = ref Cmdline.journal_name
let get_name () = !filename
let set_name s = filename := s

let print_header () =
  let time = now () in
  Format.pp_open_hvbox journal_fmt 0; (* the outermost box *)
  Format.fprintf journal_fmt
    "@[(* Frama-C journal generated at %02d:%02d the %02d/%02d/%d *)@]@;@;"
    time.Unix.tm_hour
    time.Unix.tm_min
    time.Unix.tm_mday
    (time.Unix.tm_mon+1)
    (time.Unix.tm_year + 1900);
  Format.fprintf journal_fmt (* open two boxes for start *)
    "(* Run the user commands *)@;@[<hv 2>let run () =@;@[<hv 0>"

let print_finished () = 
  (* close the outermost box *)
  Format.pp_close_box journal_fmt ()

let start () = 
  if not !something_to_write then print_header ();
  started := true;
  enable_exn := true;
  something_to_write := true

let stop () = 
  started := false;
  enable_exn := false

let clear ?(restart=true) () = 
  Buffer.clear journal_buffer; 
  something_to_write := false; 
  if restart then start ()


let preserved_files = ref []
let keep_file s = preserved_files := s :: !preserved_files

let rec get_filename =
  let cpt = ref 0 in
  let rec get_filename first =
    let name = !filename ^ ".ml" in
    if (not first && Sys.file_exists name)
      || List.mem name !preserved_files
    then begin
      incr cpt;
      let suf = "_" ^ string_of_int !cpt in
      (try 
	 let n =
	   Str.search_backward
	     (Str.regexp "_[0-9]+") 
	     !filename 
	     (String.length !filename - 1)
	 in
	 filename := Str.string_before !filename n ^ suf
       with Not_found ->
	 filename := !filename ^ suf);
      get_filename false
    end else
      name
  in
  fun () -> get_filename true

let write () =
  if !something_to_write then begin
    (* close the boxes of start *)
    Format.fprintf journal_fmt "()@]@]@;@;";
    Format.fprintf journal_fmt "@[(* Main *)@]@\n";
    Format.fprintf journal_fmt "@[<hv 2>let main () =@;";
    Format.fprintf journal_fmt
      "@[<hv 0>@[<hv 2>Journal.keep_file@;\"%s.ml\";@]@;" 
      !filename;
    Format.fprintf journal_fmt "try run ()@;";
    Format.fprintf journal_fmt "@[<hv 2>with e ->@;";
    Format.fprintf journal_fmt 
      "@[<hv 2>Format.eprintf@;\"Journal raised an exception: %%s@@.\"@;";
    Format.fprintf journal_fmt "(Printexc.to_string e)@]@]@]@]@\n@\n";
    Format.fprintf journal_fmt "@[(* Registering *)@]@\n";
    Format.fprintf journal_fmt
      "@[<hv 2>let main : unit -> unit =@;@[<hv 2>Dynamic.register@;\"%s.main\"@;"
      (String.capitalize (Filename.basename (get_name ())));
    Format.fprintf journal_fmt "@[<hv 2>(Type.func@;Type.unit@;Type.unit)@]@;";
    Format.fprintf journal_fmt "~journalize:false@;main@]@]@\n@\n";
    Format.fprintf journal_fmt "@[(* Hooking *)@]@\n";
    Format.fprintf journal_fmt "@[<hv 2>let () =@;";
    Format.fprintf journal_fmt
      "@[<hv 2>Cmdline.run_after_loading_stage@;main;@]@;";
    Format.fprintf journal_fmt "@[<hv 2>Cmdline.is_going_to_load@;()@]@]@.";
    let error msg s = error "cannot %s journal (%s)." msg s in
    let filename = get_filename () in
    feedback ~level:2 "writing journal in file \"%s\"" filename;
    try
      let cout = open_out filename in begin
	try
	  print_finished ();
	  Format.pp_print_flush journal_fmt ();
	  Buffer.output_buffer cout journal_buffer;
	with Sys_error s -> 
	  error "write into" s
      end;
      try close_out cout with Sys_error s -> error "close" s
    with Sys_error s -> 
      error "create" s
  end

let () = at_exit write

(* ****************************************************************************)
(** {2 Journalization} *)
(* ****************************************************************************)

exception Not_writable of string
let never_write name f = 
  if Cmdline.use_type then
    if Obj.tag (Obj.repr f) = Obj.closure_tag then
      Obj.magic
	(fun y -> 
	   if !started then raise (Not_writable name) else Obj.magic f y)
    else
      invalid_arg ("[Journal.never_write] " ^ name ^ " is not a closure")
  else
    f

let pp ty fmt (x:Obj.t) =
  try Type.pp ty Type.Call fmt (Obj.obj x)
  with Type.NoPrinter _ ->
    fatal
      "no printer registered for value of type %s.@.Journalisation is not possible. Aborting@." 
      (Type.name ty)

let gen_binding =
  let ids = Hashtbl.create 7 in
  let rec gen s =
    try
      let n = succ (Hashtbl.find ids s) in
      Hashtbl.replace ids s n;
      gen (s ^ "_" ^ string_of_int n)
    with Not_found ->
      Hashtbl.add ids s 1;
      s
  in
  gen

let extend_continuation f_acc pp_arg opt_label arg fmt =
  f_acc fmt;
  match opt_label with
  | None (* no label *) -> Format.fprintf fmt "@;%a" pp_arg arg;
  | Some (_, Some f) when f () == arg ->
      (* [arg] is the default value of the optional label *)
      ()
  | Some (l, _) (* other label *) -> Format.fprintf fmt "@;~%s:%a" l pp_arg arg

let catch_exn f_acc ret_ty exn fmt =
  if !enable_exn then
    Format.fprintf fmt
      "@[<hv 3>(* exception %s@;raised on: *)@]@;@[<hv 2>let __@;: %s =@;%t in@]@;"
      (Printexc.to_string exn)
      (Type.name ret_ty)
      f_acc

(* print any comment *)
let print_comment fmt pp = match pp with
  | None -> () 
  | Some pp -> Format.fprintf fmt "(* %t *)@;" pp

let print_sentence f_acc is_dyn ty comment value fmt =
  print_comment fmt comment;
  (* open a new box for the sentence *)
  Format.fprintf journal_fmt "@[<hv 2>";
  (* add a let binding whenever the return type is not unit *)
  if not (Type.equal ty Type.unit) then
    Format.fprintf journal_fmt "let %t=@;"
      (fun fmt ->
	 let binding =
	   match Type.varname ty with
	   | None -> "__" (* no binding: ignore the result *)
	   | Some f -> 
	       (* bind to a fresh variable name *)
	       let v = Obj.obj value in
	       let b = gen_binding (f v) in
	       Type.Binding.add ty v b;
	       b
	 in
	 Format.fprintf fmt 
	   "%s%s" 
	   binding
	   (* add the return type for dynamic application *)
	   (if is_dyn then "@;: " ^ Type.name ty else " "));
  (* pretty print the sentence itself in a box *)
  Format.fprintf fmt "@[<hv 2>%t@]" f_acc;
  (* close the sentence *)
  if Type.equal ty Type.unit then Format.fprintf fmt ";@]@;" 
  else Format.fprintf journal_fmt "@;<1 -2>in@]@;"

let rec journalize_function f_acc ty is_dyn comment (x:Obj.t) =
  if Type.Function.is_instance_of ty then begin
    (* [ty] is a function type value:
       there exists [a] and [b] such than [ty = a -> b] *)
    let ty : ('a,'b) Type.Function.poly Type.t = Obj.magic (ty:'ty Type.t) in
    let (a:'a Type.t), (b:'b Type.t), opt_label = 
      Type.Function.get_instance ty 
    in
    Obj.repr
      (fun (y:'a) ->
	 if not !started (*|| !running*) then
	   (* prevent journalisation if you're journalizing another function *)
	   Obj.repr (Obj.obj x y)
	 else
	   try
	     (* [started] prevents journalization of function call 
		inside another one *)
	     let old_started = !started in
	     started := false;
	     (* apply the closure [x] to its argument [y] *)
	     let xy = Obj.obj x y in
	     started := old_started;
	     (* extend the continuation and continue *)
	     let f_acc = extend_continuation f_acc (pp a) opt_label y in
	     journalize_function f_acc b is_dyn comment xy
	   with 
	     | Not_writable name ->
		 fatal
		   "a call to the function \"%s\" cannot be written in the journal"
		   name
	     | exn ->
		 if !enable_exn then catch_exn f_acc b exn journal_fmt;
		 raise exn)
  end else begin
    if !started then print_sentence f_acc is_dyn ty comment x journal_fmt;
    x
  end
      
let register s ty ?comment ?(is_dyn=false) x =
  if Cmdline.journal_enable then begin
    if s = "" then 
      abort "[Journal.register] the given name should not be \"\"";
    Type.Binding.add_once ty x s;
    if Type.Function.is_instance_of ty then begin
      let x' = Obj.repr x in
      let f_acc fmt = pp ty fmt x' in
      let res : Obj.t = journalize_function f_acc ty is_dyn comment x' in
      Obj.obj res
    end else
      x
  end else
    x

(*
Local Variables:
compile-command: "LC_ALL=C make -C ../.. -j"
End:
*)
