(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2009                                               *)
(*    CEA (Commissariat  l'nergie Atomique)                             *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(* --- Logging Facilities                                                 --- *)
(* -------------------------------------------------------------------------- *)

open Format

let kid = ref 0  (* channel ref count *)
let min_buffer = 80    (* initial size of buffer *)
let max_buffer = 4096  (* buffer resized after useage if larger than max *)

exception AbortError of string (* channel name *)
exception AbortFatal of string (* channel name *)

type terminal = {
  mutable gid : int ;   (* last channel that writes on *)
  mutable cut : bool ;  (* last channel was cut *)
  out : out_channel ;
}

let stdout = {
  gid = (-1) ;
  cut = false ;
  out = Pervasives.stdout ;
}

let stderr = {
  gid = (-1) ;
  cut = false ;
  out = Pervasives.stderr ;
}

let null = Format.make_formatter (fun _ _ _ -> ()) (fun _ -> ())

type kind = Result | Feedback | Debug | Warning | Error | Fatal
type source = { src_file : string ; src_line : int }

(*
let source (loc,_) = { 
  src_file = loc.Lexing.pos_fname ; 
  src_line = loc.Lexing.pos_lnum ;
}
*)

let current_loc = ref (fun () -> raise Exit)
let set_current_source cloc = current_loc := cloc

type event = {
  evt_kind : kind ;
  evt_source : source option ;
  evt_channel : string ;
  evt_message : string ;
}

type channel = {
  mutable buffer : string ;
  mutable pos : int ; (* end of message in buffer *)
  mutable last_line : int ; (* beginning of last line (partially) echoed.  *)
  mutable last_char : int ; (* position of last char echoed. *)
  mutable formatter : formatter ;
  mutable listeners : (event -> unit) list ;
  mutable source : source option ;
  mutable continued : bool ;
  mutable echo : bool ;
  mutable emitter : (event -> unit) option ;
  mutable once : bool ;
  oncetable : (source option * string,unit) Hashtbl.t ;
  kind : kind ;
  channel : string ;
  id : int ;
  autoflush: bool;
  label : string ;
  term : terminal ;
}

type listener = string option * kind option * (event -> unit)
type echo = string option * kind option * bool

let channels : channel list ref = ref []
let listeners : listener list ref = ref []
let echoes : echo list ref = ref []

let do_clear c =
  begin
    c.source <- None ;
    c.emitter <- None ;
    c.pos <- 0 ;
    c.last_line <- 0 ;
    c.last_char <- 0 ;
    c.once <- false ;
    c.continued <- false ;
    if String.length c.buffer > max_buffer then
      c.buffer <- String.create min_buffer
  end
  
(* -------------------------------------------------------------------------- *)
(* --- Listeners                                                          --- *)
(* -------------------------------------------------------------------------- *)

let channel_matches c channel kind =
  ( match channel with None -> true | Some p -> p = c.channel ) &&
    ( match kind with None -> true | Some k -> k = c.kind )

let event_matches e channel kind =
  ( match channel with None -> true | Some p -> p = e.evt_channel ) &&
    ( match kind with None -> true | Some k -> k = e.evt_kind )
    
let apply_echo c (channel,kind,active) =
  if channel_matches c channel kind then c.echo <- active

let apply_listener c (channel,kind,demon) =
  if channel_matches c channel kind then c.listeners <- c.listeners @ [ demon ]

let notify e =
  List.iter
    (fun (channel,kind,demon) ->
       if event_matches e channel kind then demon e
    ) !listeners

let get_event c = {
  evt_source = c.source ;
  evt_message = String.sub c.buffer 0 c.pos ;
  evt_channel = c.channel ;
  evt_kind = c.kind ;
} 

let do_fire c evt =
  match c.emitter with
    | Some p -> p evt
    | None ->
	match c.listeners with
	  | [] -> ()
	  | demons ->
	      List.iter (fun p -> try p evt with _ -> ()) demons
    
(* -------------------------------------------------------------------------- *)
(* --- Flushing Output                                                    --- *)
(* -------------------------------------------------------------------------- *)
  
let echo_substring c label k n =
  begin
    if label then
      Pervasives.output c.term.out c.label 0 (String.length c.label) ;
    Pervasives.output c.term.out c.buffer k n ;
    Pervasives.flush c.term.out ;
  end

let rec echo_lines c label p =
  if label && p = 0 then
    begin
      match c.source with
	| Some src -> 
	    (* No label here for compatibility with find-next-error 
	       of Emacs mode. *)
	    Printf.fprintf c.term.out "File \"%s\", line %d:\n" 
	      src.src_file src.src_line
	| None -> ()
    end ;
  if p >= c.pos then
    begin
      c.last_char <- c.pos ;
      c.last_line <- c.pos ;
      c.term.cut <- false ;
    end
  else
    let q = try String.index_from c.buffer p '\n' with Not_found -> (-1) in
    if q < 0 || q >= c.pos then
      begin
	let r = c.pos - p in
	echo_substring c label p r ;
	if label then c.last_line <- p ;
	c.last_char <- c.pos ;
	c.term.cut <- true ;
      end
    else
      begin
	echo_substring c label p (1+q-p) ;
	echo_lines c true (q+1) ;
      end

let do_echo c =
  if c.echo && c.emitter = None && not c.once && c.pos > c.last_char 
  then
    if c.term.gid = c.id then
      echo_lines c (c.last_line = c.last_char) c.last_char
    else
      begin
	if c.term.cut then
	  ( Pervasives.output c.term.out " [...]\n" 0 7 ;
	    Pervasives.flush c.term.out ) ;
	c.term.gid <- c.id ;
	echo_lines c true c.last_line ;
      end
	
let channel_flush c () = do_echo c

(* -------------------------------------------------------------------------- *)
(* --- Buffering Output                                                   --- *)
(* -------------------------------------------------------------------------- *)

let rec size_up required size =
  let s = 2*size+1 in
  if required <= s then s else size_up required s

let append c text k n =
  begin
    let req = c.pos + n in
    let avail = String.length c.buffer in
    if req > avail then
      begin
	let s = size_up req avail in
	let t = String.create s in
	String.blit c.buffer 0 t 0 c.pos ;
	c.buffer <- t ;
      end ;
    String.blit text k c.buffer c.pos n ;
    c.pos <- c.pos + n ;
  end

let channel_out c text k n =
  if n > 0 then
    begin
      append c text k n ;
      if c.autoflush then
	let has_cr = 
	  ( n=1 && text.[k]='\n' ) ||
	  ( try (String.index_from text k '\n') < k+n with Not_found -> false )
	in if has_cr then do_echo c ;
    end

(* -------------------------------------------------------------------------- *)
(* --- Management & Utilities                                             --- *)
(* -------------------------------------------------------------------------- *)

let new_channel ~channel ~kind ~label ~autoflush ~term =
  begin
    incr kid ;
    let c = {
      buffer = String.create min_buffer ; 
      pos = 0 ; last_char = 0 ; last_line = 0 ;
      id = !kid ; label = label ; autoflush = autoflush;
      channel = channel ; kind = kind ; 
      source = None ; emitter = None ;
      continued = false ; echo = true ; once = false ;
      formatter = Format.err_formatter ;
      listeners = [] ; term = term ;
      oncetable = Hashtbl.create 71 ;
    } in
    c.formatter <- Format.make_formatter (channel_out c) (channel_flush c) ;
    List.iter (apply_listener c) !listeners ;
    List.iter (apply_echo c) !echoes ;
    (* order is important *)
    channels := !channels @ [c] ; c
  end

let add_listener ?channel ?kind demon =
  begin
    let listener = channel,kind,demon in
    List.iter (fun c -> apply_listener c listener) !channels ;
    (* order is important *)
    listeners := !listeners @ [listener] ;
  end

let set_echo ?channel ?kind active =
  begin
    let echo = channel,kind,active in
    List.iter (fun c -> apply_echo c echo) !channels ;
    (* order is important *)
    echoes := !echoes @ [echo] ;
  end

let finally_emit c fmt =
  begin
    pp_print_flush fmt () ;
    if c.pos > 0 && c.buffer.[pred c.pos] <> '\n' then
      append c "\n" 0 1 ;
    if c.once then
      begin
	let evt = get_event c in
	let m = evt.evt_source , evt.evt_message in
	if not (Hashtbl.mem c.oncetable m) then
	  ( Hashtbl.add c.oncetable m () ;
	    c.once <- false ;
	    do_echo c ; 
	    do_fire c evt )
	    (* else nothing ! *)
      end
    else
      begin
	do_echo c ; 
	do_fire c (get_event c) ;
      end ;
    do_clear c ;
  end

let finally_raise c e fmt =
  begin
    finally_emit c fmt ;
    raise e ;
  end

let do_terminate c = finally_emit c c.formatter

let fprintf c =
  let fmt = c.formatter in
  fun msg -> Format.fprintf fmt msg

type 'a pretty_printer = 
    ?current:bool -> ?source:source -> 
    ?continue:bool -> ?last:bool -> 
    ?emitwith:(event -> unit) -> ?once:bool ->
    ('a,formatter,unit) format -> 'a
    
type ('a,'b) pretty_aborter =
    ?current:bool -> ?source:source -> 
    ('a,formatter,unit,'b) format4 -> 'a

type pretty_formatter = 
    ?current:bool -> ?source:source ->
    ?emitwith:(event -> unit) -> ?once:bool ->
    unit -> formatter
    
(* It is useless to define optional parameter here, since
   generalization of polymorphic variables can not be done before
   use => necessity to redefine (let f x = g x is different from let f = g) 
*)
    
let do_flow c current source continue last emitter once =
  if c.continued && not continue && not last then do_terminate c ;
  c.continued <- continue && not last ;
  if c.pos = 0 then
    begin
      c.source <- 
	if current then (try Some( !current_loc () ) with Exit -> None)
	else source ;
      c.emitter <- emitter ;
      c.once <- once ;
    end
    
let fprintf_emit c =
  let fmt = c.formatter in
  fun current source continue last emitter once msg ->
    Format.pp_print_flush fmt () ;
    do_flow c current source continue last emitter once ;
    if c.continued
    then Format.fprintf fmt msg
    else Format.kfprintf (finally_emit c) fmt msg

let fprintf_abort c s =
  let fmt = c.formatter in
  fun current source msg ->
    do_flow c current source false false None false ;
    Format.kfprintf (finally_raise c (AbortError s)) fmt msg
      
let fprintf_fatal c s =
  let fmt = c.formatter in
  fun current source msg ->
    do_flow c current source false false None false ;
    Format.kfprintf (finally_raise c (AbortFatal s)) fmt msg
  

let fprintf_log c verb =
  let fmt = c.formatter in
  fun level current source continue last emitter once msg ->
    do_flow c current source continue last emitter once ;
    if verb level then
      if c.continued
      then Format.fprintf c.formatter msg
      else Format.kfprintf (finally_emit c) c.formatter msg
    else
      Format.ifprintf fmt msg

let fprintf_assert c e =
  fun current source msg ->
    do_flow c current source false false None false ;
    let fmt = if e then null else c.formatter in
    Format.kfprintf (fun fmt -> finally_emit c fmt ; e) fmt msg

let fmt_get c current source emitter once =
  do_flow c current source true false emitter once ; c.formatter

let fmt_log c verb level current source emitter once =
  do_flow c current source true false emitter once ;
  if verb level then c.formatter else null

let pretty_printer c =
  fun ?(current=false) ?source ?(continue=false) ?(last=false) 
    ?emitwith ?(once=false) msg ->
    fprintf_emit c current source continue last emitwith once msg

(* -------------------------------------------------------------------------- *)
(* --- Plugin Interface                                                   --- *)
(* -------------------------------------------------------------------------- *)

let kernel_channel_name = "kernel"
let kernel_label_name = "Frama-C"

let echo e =
  try
    let c =
      List.find 
	(fun c -> c.channel = e.evt_channel && c.kind = e.evt_kind)
	!channels
    in 
    do_terminate c ; 
    append c e.evt_message 0 (String.length e.evt_message) ;
    do_echo c ;
    do_clear c ;
  with Not_found ->
    let k = match e.evt_kind with 
      | Result -> "result"
      | Feedback -> "feedback"
      | Debug -> "debug"
      | Warning -> "warning"
      | Error -> "error"
      | Fatal -> "fatal"
    in
    let src = match e.evt_source with
      | None -> ""
      | Some s ->
	  Format.sprintf "File \"%s\", line %d:\n" s.src_file s.src_line in
    let msg = 
      Format.sprintf "Message echoed on non-existent channel [%s:%s]:%s\n%s"
	e.evt_channel k src e.evt_message 
    in failwith msg

module type Messages =
sig

  val verbose_atleast : int -> bool
  val debug_atleast : int -> bool

  val result  : ?level:int -> 'a pretty_printer
  val feedback: ?level:int -> 'a pretty_printer
  val debug   : ?level:int -> 'a pretty_printer
  val warning : 'a pretty_printer
  val error   : 'a pretty_printer
  val abort   : ('a,'b) pretty_aborter
  val fatal   : ('a,'b) pretty_aborter
  val verify  : bool -> ('a,bool) pretty_aborter

  val result_fmt   : ?level:int -> pretty_formatter
  val feedback_fmt : ?level:int -> pretty_formatter
  val debug_fmt    : ?level:int -> pretty_formatter
  val warning_fmt  : pretty_formatter
  val error_fmt    : pretty_formatter
  val fatal_fmt    : pretty_formatter

end

module Register
  (P : sig
     val channel : string
     val label : string
     val verbose_atleast : int -> bool
     val debug_atleast : int -> bool
   end) =
struct

  include P

  let new_chan k fmt autoflush term =
    new_channel P.label k (Printf.sprintf fmt P.channel) autoflush term

  let c_result  = new_chan Result   "[%s] "          false  stdout
  let c_debug   = new_chan Debug    "[%s (debug)] "  true   stderr
  let c_feedback= new_chan Feedback "[%s...] "       true   stderr
  let c_warning = new_chan Warning  "[%s] warning: " true   stdout
  let c_error   = new_chan Error    "[%s] error: "   true   stdout
  let c_fatal   = new_chan Fatal    "[%s] internal error: " true stdout

  let result
      ?(level=1) ?(current=false) ?source 
      ?(continue=false) ?(last=false) ?emitwith ?(once=false) msg = 
    fprintf_log c_result verbose_atleast level
      current source continue last emitwith once msg
  let feedback ?(level=1) ?(current=false) ?source 
      ?(continue=false) ?(last=false) ?emitwith ?(once=false) msg = 
    fprintf_log c_feedback verbose_atleast level
      current source continue last emitwith once msg
  let debug  ?(level=1) ?(current=false) ?source 
      ?(continue=false) ?(last=false) ?emitwith ?(once=false) msg = 
    fprintf_log c_debug  debug_atleast level 
      current source continue last emitwith once msg
  let warning ?(current=false) ?source 
      ?(continue=false) ?(last=false) ?emitwith ?(once=false) msg = 
    fprintf_emit c_warning 
      current source continue last emitwith once msg
  let error ?(current=false) ?source
      ?(continue=false) ?(last=false) ?emitwith ?(once=false) msg = 
    fprintf_emit c_error 
      current source continue last emitwith once msg
  let abort ?(current=false) ?source msg = 
    fprintf_abort c_error P.channel current source msg
  let fatal ?(current=false) ?source msg = 
    fprintf_fatal c_fatal P.channel current source msg

  let verify assertion ?(current=false) ?source msg =
    fprintf_assert c_fatal assertion current source msg

  let result_fmt ?(level=1) ?(current=false) ?source ?emitwith ?(once=false) () =
    fmt_log c_result verbose_atleast level current source emitwith once
  let feedback_fmt ?(level=1) ?(current=false) ?source ?emitwith ?(once=false) () =
    fmt_log c_feedback verbose_atleast level current source emitwith once
  let debug_fmt ?(level=1) ?(current=false) ?source ?emitwith ?(once=false) () =
    fmt_log c_debug debug_atleast level current source emitwith once
  let warning_fmt ?(current=false) ?source ?emitwith ?(once=false) () =
    fmt_get c_debug current source emitwith once
  let error_fmt ?(current=false) ?source ?emitwith ?(once=false) () =
    fmt_get c_error current source emitwith once
  let fatal_fmt ?(current=false) ?source ?emitwith ?(once=false) () =
    fmt_get c_fatal current source emitwith once

end

(* -------------------------------------------------------------------------- *)
(* --- Termination                                                        --- *)
(* -------------------------------------------------------------------------- *)

let flush_all () = 
  List.iter do_terminate !channels

let () = at_exit flush_all


(* -------------------------------------------------------------------------- *)
(*
Local Variables:
compile-command: "LC_ALL=C make -C ../.."
End:
*)
