(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2009                                               *)
(*    CEA (Commissariat  l'nergie Atomique)                             *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(** Logging Services for Frama-C Kernel and Plugins.
    This is a {i low-level} interface. Not to be used by casual users.
    @since Beryllium-20090601-beta1 *)
(* -------------------------------------------------------------------------- *)

open Format

exception AbortError of string
  (** User error that prevents a plugin to terminate. 
      @since Beryllium-20090601-beta1 *)

exception AbortFatal of string
  (** Internal error that prevents a plugin to terminate. 
      @since Beryllium-20090601-beta1 *)

type terminal
  (** @since Beryllium-20090601-beta1 *)

val stdout : terminal
  (** @since Beryllium-20090601-beta1 *)

val stderr : terminal
  (** @since Beryllium-20090601-beta1 *)

type kind = Result | Feedback | Debug | Warning | Error | Fatal
  (** @since Beryllium-20090601-beta1 *)

type source = { src_file : string ; src_line : int }
  (** @since Beryllium-20090601-beta1 *)

val set_current_source : (unit -> source) -> unit
  (** @since Beryllium-20090601-beta1 *)

type event = {
  evt_kind : kind ;
  evt_source : source option ;
  evt_channel : string ;
  evt_message : string ;
}
  (** @since Beryllium-20090601-beta1 *)

type channel
  (** @since Beryllium-20090601-beta1 *)

val new_channel : channel:string -> kind:kind -> label:string ->
  autoflush:bool -> term:terminal -> channel
  (** @since Beryllium-20090601-beta1 *)

type 'a pretty_printer = 
    ?current:bool -> ?source:source ->
  ?continue:bool -> ?last:bool ->     
  ?emitwith:(event -> unit) -> ?once:bool ->
  ('a,formatter,unit) format -> 'a 
  (** @since Beryllium-20090601-beta1 *)

(** A pretty printer manages the emission of {i messages}.
    
    A {i message} is a string procuced by one or several successive calls to
    [fprintf] variants provided by the [Log] module.
    
    Messages are echoed on standard outputs of the application (Cf. [stdout]
    and [stderr]). The [Log] module manages interleaving between the
    different channels.
    
    Additionaly, messages are emitted on event listeners.

    A message can be {i continued} over several calls to pretty
    printer with the option [~continue:true] parameter, the emission
    of events is delayed until the message is complete. However,
    echo on standard output may be still immediately flushed for
    direct feedback to the user (this is typicaly the case for error
    channels).

    To complete a message, simply call the printer with an empty
    string or use the [~last:true] optional argument. Otherwised,
    the message would be implicitely terminated when the next new
    message is produced on the same channel.
    
    When a source location is present, it must be present from the
    very first emission of the message, otherwised it may not be
    echoed correctly. It is also possible to use [~current:true] for
    the current location [Cil.CurrentLoc]. 

    The optional parameter [~once:true] can be set at the beginning
    of a new message to emit it only once. As a side effect, echo on
    standard output is delayed until the message is complete. The
    message is thrown away if already emitted. If a source location
    is specified, the location is taken into account for registering
    the message.

    When the optional parameter [~emitwith:p] is used, echo on
    standard output {i and} notification to listener are both
    disabled. The message is still build and passed to the callback
    function [p] instead. It can be later echoed with [echo] or
    notified with [notify]. *)
    
type ('a,'b) pretty_aborter =
    ?current:bool -> ?source:source -> 
    ('a,formatter,unit,'b) format4 -> 'a
  (** Variant for pretty printers that raise an exception. 
	@since Beryllium-20090601-beta1 *)

type pretty_formatter = 
    ?current:bool -> ?source:source -> 
  ?emitwith:(event -> unit) -> ?once:bool ->
  unit -> formatter
    (** Variant for obtaining the formatters. 
	The flag [~continue:true] is positionned by default.

	To terminate the message, you must invoke the associated
	pretty-printer with [~continue:false] or [~last:true]
	flags. Flushing the formatter {i do} flush the echo on
	standard outputs, but {i do not} terminate the message. 
	@since Beryllium-20090601-beta1 *)
  
(** {2 Pretty Printing} *)

val fprintf : channel -> ('a,formatter,unit) format -> 'a
  (** @since Beryllium-20090601-beta1 *)

val pretty_printer : channel -> 'a pretty_printer
  (** @since Beryllium-20090601-beta1 *)

(** {2 Echo and Notification} *)

val set_echo : ?channel:string -> ?kind:kind -> bool -> unit
  (** @since Beryllium-20090601-beta1 *)

val add_listener : ?channel:string -> ?kind:kind -> (event -> unit) -> unit
  (** @since Beryllium-20090601-beta1 *)

val echo : event -> unit
  (** @since Beryllium-20090601-beta1 *)

val notify : event -> unit
  (** @since Beryllium-20090601-beta1 *)

val flush_all : unit -> unit
  (** Flush all existing [channel]. It is called [at_exit]. 
      @since Beryllium-20090601-beta1 *)

(** {2 Plugin Interface} *)

val kernel_channel_name: string
  (** the reserved channel name used by the Frama-C kernel. 
      @since Beryllium-20090601-beta1 *)

val kernel_label_name: string
  (** the reserved label name used by the Frama-C kernel. 
      @since Beryllium-20090601-beta1 *)

(** @since Beryllium-20090601-beta1 *)
module type Messages = sig

  val verbose_atleast : int -> bool
    (** @since Beryllium-20090601-beta1 *)

  val debug_atleast : int -> bool
    (** @since Beryllium-20090601-beta1 *)

  val result  : ?level:int -> 'a pretty_printer
    (** Results of analysis. Default level is 1. 
	@since Beryllium-20090601-beta1 *)

  val feedback : ?level:int -> 'a pretty_printer
    (** Progress and feedback. Level is tested against the verbose. 
	@since Beryllium-20090601-beta1 *)

  val debug   : ?level:int -> 'a pretty_printer
    (** Debugging information dedicated to Plugin developpers. 
	Default level is 1. 
	@since Beryllium-20090601-beta1 *)

  val warning : 'a pretty_printer
    (** Hypothesis and restrictions. 
	@since Beryllium-20090601-beta1 *)

  val error   : 'a pretty_printer
    (** user error: syntax/typing error, bad expected input, etc.
	@since Beryllium-20090601-beta1 *)

  val abort   : ('a,'b) pretty_aborter
    (** user error stopping the plugin.
        @raise AbortError with the channel name. 
	@since Beryllium-20090601-beta1 *)

  val fatal   : ('a,'b) pretty_aborter
    (** internal error of the plug-in. 
        @raise AbortFatal with the channel name. 
	@since Beryllium-20090601-beta1 *)

  val verify : bool -> ('a,bool) pretty_aborter
    (** If the first argument is [true], return [true] and do nothing else,
	otherwise, send the message on the {i fatal} channel and return
	[false].  
	
	The intended usage is: [assert (verify e "Bla...") ;]. 
	@since Beryllium-20090601-beta1 *)

  val result_fmt : ?level:int -> pretty_formatter
    (** Formatter to continue or initiates a message on the [result] channel. 
	@since Beryllium-20090601-beta1 *)

  val feedback_fmt : ?level:int -> pretty_formatter
    (** Formatter to continue or initiate a message on the [feedback] channel.
	@since Beryllium-20090601-beta1 *)

  val debug_fmt : ?level:int -> pretty_formatter
    (** Formatter to continue or initiate a message on the [debug] channel. 
	@since Beryllium-20090601-beta1 *)

  val warning_fmt : pretty_formatter
    (** Formatter to continue or initiate a [warning] message. 
	@since Beryllium-20090601-beta1 *)

  val error_fmt : pretty_formatter
    (** Formatter to continue or initiate an [error] message. 
	@since Beryllium-20090601-beta1 *)

  val fatal_fmt : pretty_formatter
    (** Formatter to continue or initiate a [fatal] error message. 
	@since Beryllium-20090601-beta1 *)

end

(** Each plugin has its own channel to output messages.
    This functor should not be directly applied by plug-in developer.
    They should apply {!Plugin.Register} instead.
    @since Beryllium-20090601-beta1 *)
module Register
  (P : sig
     val channel : string
     val label : string
     val verbose_atleast : int -> bool
     val debug_atleast : int -> bool
   end) 
  : Messages

(*
Local Variables:
compile-command: "LC_ALL=C make -C ../.."
End:
*)
